import ast
from typing import List

from flake8_plugin_utils import Error

from flake8_qa_style.checkers.node_visitors.function_call_visitor import (
    Checker,
    Context,
    FunctionCallVisitor,
)
from flake8_qa_style.errors import SleepWithConstantArgument


@FunctionCallVisitor.register_checker
class SleepChecker(Checker):

    def check(self, call_node: ast.Call, context: Context) -> List[Error]:
        if (
            context.sleep_function_name
            and isinstance(call_node.func, ast.Name)
            and call_node.func.id == context.sleep_function_name
        ) or (
            context.is_imported_module_time
            and isinstance(call_node.func, ast.Attribute)
            and call_node.func.attr == 'sleep'
            and call_node.func.value.id == 'time'
        ):
            if call_node.args:
                if isinstance(call_node.args[0], ast.Constant):
                    return [SleepWithConstantArgument(call_node.lineno, call_node.col_offset)]
            else:
                if (
                    call_node.keywords
                    and call_node.keywords[0].arg == 'secs'
                    and isinstance(call_node.keywords[0].value, ast.Constant)
                ):
                    return [SleepWithConstantArgument(call_node.lineno, call_node.col_offset)]

        return []



from flake8_qa_style.checkers.raw_checkers._raw_checker import RawChecker
from flake8_qa_style.errors import NoBlankFileStart


class FileStartChecker(RawChecker):
    def check(self) -> None:
        if not self.lines:
            return

        if not self.lines[0].strip():
            self.errors.append(NoBlankFileStart(lineno=1, col_offset=0))



from flake8_plugin_utils import assert_error

from flake8_qa_style.checkers.node_visitors import FunctionCallVisitor
from flake8_qa_style.checkers.node_visitors.call_checkers import PrintChecker
from flake8_qa_style.errors import Print


def test_call_print():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(PrintChecker)
    code = """
    print('asdasd')
    """
    assert_error(FunctionCallVisitor, code, Print)


def test_call_pp():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(PrintChecker)
    code = """
    pp('asdasd')
    """
    assert_error(FunctionCallVisitor, code, Print)



import ast
import textwrap
from typing import Any, Optional, Type

from flake8_plugin_utils.plugin import Error, TConfig, Visitor


def _errors_from_src(
    visitor_cls: Type[Visitor[TConfig]],
    src: str,
    config: Optional[TConfig] = None
) -> Optional[list[Error]]:
    visitor = visitor_cls(config=config)
    tree = ast.parse(textwrap.dedent(src))
    visitor.visit(tree)

    if not visitor.errors:
        return None

    return visitor.errors


def assert_errors(
    visitor_cls: Type[Visitor[TConfig]],
    src: str,
    expected: list[Type[Error]],
    config: Optional[TConfig] = None,
    **kwargs: Any,
) -> None:
    errors = _errors_from_src(visitor_cls, src, config=config)
    assert errors, f'Errors "{errors}" not found in\n{src}'
    assert len(errors) == len(expected)

    for error, expected_error in zip(errors, expected):
        assert isinstance(error, expected_error)

        expected_message = expected_error.formatted_message(**kwargs)
        assert expected_message == error.message, (
            f'Expected error with message "{expected_message}", got "{error.message}"'
        )



class Config:
    def __init__(self, skip_property_return_annotation: bool):
        self.skip_property_return_annotation = skip_property_return_annotation


class DefaultConfig(Config):
    def __init__(self, skip_property_return_annotation: bool = False):
        super().__init__(skip_property_return_annotation=skip_property_return_annotation)



from flake8_plugin_utils import Error


class SleepWithConstantArgument(Error):
    code = 'CS001'
    message = 'using sleep with constant as argument is prohibited. Use sleep(var) instead'


class Print(Error):
    code = 'CS002'
    message = 'try to avoid printing. If it\'s necessary ignore this warning'


# assert var == var
class AssertSameObjectsForEquality(Error):
    code = 'CS003'
    message = 'comparing same objects in assert doesn\'t make any sense'


# assert var == 'string'
# assert var == 2
class AssertWithConstant(Error):
    code = 'CS004'
    message = 'assert variable with constants is not allowed, move constant to libs'


class ArgAnnotationMissing(Error):
    code = 'CS005'
    message = 'missing annotation for argument "{arg_name}" for function "{func_name}"'


class ReturnAnnotationMissing(Error):
    code = 'CS006'
    message = 'missing return annotation for function "{func_name}"'


class NoBlankFileStart(Error):
    code = 'CS007'
    message = 'file should not start with a blank line'


class NestedWith(Error):
    code = 'CS008'
    message = 'nested "with" statements of the same type should be comma-separated'



from flake8_plugin_utils import assert_error, assert_not_error

from flake8_qa_style.checkers.node_visitors import AnnotationVisitor
from flake8_qa_style.config import DefaultConfig
from flake8_qa_style.errors import ArgAnnotationMissing, ReturnAnnotationMissing


def test_func_no_annotation():
    code = """
    def f(var): pass
    """
    assert_error(AnnotationVisitor, code, ArgAnnotationMissing, arg_name='var', func_name='f',
                 config=DefaultConfig())


def test_func_with_annotation():
    code = """
    def f(var: int): pass
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_func_with_partial_annotation():
    code = """
    def f(var_1: int, var_2): pass
    """
    assert_error(AnnotationVisitor, code, ArgAnnotationMissing, arg_name='var_2', func_name='f',
                 config=DefaultConfig())


def test_func_with_annotation_and_defaults():
    code = """
    def f(var_1: int, var_2: bool = True): pass
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_func_without_annotation_and_defaults():
    code = """
    def f(var_1: int, var_2 = True): pass
    """
    assert_error(AnnotationVisitor, code, ArgAnnotationMissing,
                 func_name='f', arg_name='var_2', config=DefaultConfig())


def test_func_self_arg():
    code = """
    def f(self): pass
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_func_cls_arg():
    code = """
    def f(cls): pass
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_func_no_return_annotation():
    code = """
    def f(): return True
    """
    assert_error(AnnotationVisitor, code, ReturnAnnotationMissing, func_name='f',
                 config=DefaultConfig())


def test_func_return_annotation():
    code = """
    def f() -> bool: return True
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_func_no_return_annotation_for_property_skipped():
    code = """
    @property
    def f(): return True
    """
    assert_not_error(AnnotationVisitor, code,
                     config=DefaultConfig(skip_property_return_annotation=True))


def test_func_no_return_annotation_for_property():
    code = """
    @property
    def f(): return True
    """
    assert_error(AnnotationVisitor, code, ReturnAnnotationMissing,
                 func_name='f',
                 config=DefaultConfig(skip_property_return_annotation=False))


def test_dunder_func_without_annotation():
    code = """
    def __init__(self, attribute):
      self.attribute = attribute
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())


def test_dunder_return_func_without_annotation():
    code = """
    def __dict__(self):
      return 'some string'
    """
    assert_not_error(AnnotationVisitor, code, config=DefaultConfig())



from flake8_plugin_utils import assert_error, assert_not_error

from flake8_qa_style.checkers.node_visitors import AssertVisitor
from flake8_qa_style.errors import AssertSameObjectsForEquality, AssertWithConstant


def test_assert_same_constants():
    code = """
    assert 1 == 1
    """
    assert_error(AssertVisitor, code, AssertSameObjectsForEquality)


def test_assert_constants():
    code = """
    assert 1 == 2
    """
    assert_not_error(AssertVisitor, code)


def test_assert_same_variables():
    code = """
    assert var == var
    """
    assert_error(AssertVisitor, code, AssertSameObjectsForEquality)


def test_assert_different_variables():
    code = """
    assert var == foo
    """
    assert_not_error(AssertVisitor, code)


def test_assert_with_string():
    code = """
    assert var == 'string'
    """
    assert_error(AssertVisitor, code, AssertWithConstant)


def test_assert_with_number():
    code = """
    assert var == 2
    """
    assert_error(AssertVisitor, code, AssertWithConstant)


def test_assert_with_string_in_variable():
    code = """
    foo = 'string'
    assert var == foo
    """
    assert_not_error(AssertVisitor, code)


def test_assert_with_fstring():
    code = """
    assert var == f'string'
    """
    assert_not_error(AssertVisitor, code)


def test_assert_in_scenario():
    code = """
    class Scenario:
        def when(): pass
        def then(): assert self.foo == self.foo
    """
    assert_error(AssertVisitor, code, AssertSameObjectsForEquality)


def test_correct_assert_in_scenario():
    code = """
    class Scenario:
        def when(): pass
        def then(): assert self.foo == self.var
    """
    assert_not_error(AssertVisitor, code)


def test_assert_with_bool():
    code = """
    assert True
    """
    assert_not_error(AssertVisitor, code)



from typing import Optional

from flake8_plugin_utils import Visitor

from flake8_qa_style.config import Config


class NodeVisitor(Visitor):
    def __init__(self, config: Optional[Config] = None, filename: Optional[str] = None):
        super().__init__(config=config)
        self.filename: Optional[str] = filename



from abc import ABC, abstractmethod
from typing import Optional

from flake8_plugin_utils import Error


class RawChecker(ABC):
    def __init__(self, filename: Optional[str] = None, lines: Optional[list[str]] = None):
        self.filename: Optional[str] = filename
        self.lines: Optional[list[str]] = lines
        self.errors: list[Error] = []

    @abstractmethod
    def check(self) -> None:
        pass



from flake8_plugin_utils import assert_error, assert_not_error

from flake8_qa_style.checkers.node_visitors import FunctionCallVisitor
from flake8_qa_style.checkers.node_visitors.call_checkers import SleepChecker
from flake8_qa_style.errors import SleepWithConstantArgument


def test_call_sleep():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    from time import sleep
    sleep(1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_no_import():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    def sleep(foo): pass
    sleep(1)
    """
    assert_not_error(FunctionCallVisitor, code)


def test_call_sleep_in_func():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    from time import sleep
    def foo(): sleep(1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_in_class():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    from time import sleep
    class Test:
        def foo(): sleep(1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_with_var():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    from time import sleep
    sleep(var)
    """
    assert_not_error(FunctionCallVisitor, code)


def test_call_sleep_as_name():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    from time import sleep as s
    s(1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_import_time():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    import time
    def f(): time.sleep(1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_import_time_kwargs():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    import time
    def f(): time.sleep(secs=1)
    """
    assert_error(FunctionCallVisitor, code, SleepWithConstantArgument)


def test_call_sleep_import_time_kwargs_not_constant():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    import time
    def f(): time.sleep(secs=a)
    """
    assert_not_error(FunctionCallVisitor, code)


def test_call_not_sleep_import_time():
    FunctionCallVisitor.deregister_all()
    FunctionCallVisitor.register_checker(SleepChecker)
    code = """
    import time
    def f(): time.not_sleep(1)
    """
    assert_not_error(FunctionCallVisitor, code)



from flake8_qa_style.checkers.raw_checkers import FileStartChecker
from flake8_qa_style.errors import NoBlankFileStart


def test_empty_file() -> None:
    checker = FileStartChecker(None, [])
    checker.check()
    assert len(checker.errors) == 0


def test_file_with_blank_line() -> None:
    checker = FileStartChecker(None, ["\n"])
    checker.check()
    assert len(checker.errors) == 1
    assert isinstance(checker.errors[0], NoBlankFileStart)


def test_file_without_blank_line() -> None:
    checker = FileStartChecker(None, ["print('Hello world')"])
    checker.check()
    assert len(checker.errors) == 0



import argparse
import ast
from typing import Callable, List, Optional

from flake8.options.manager import OptionManager
from flake8_plugin_utils import Plugin, Visitor

from flake8_qa_style.checkers.node_visitors import (
    AnnotationVisitor,
    AssertVisitor,
    FunctionCallVisitor,
    WithVisitor,
)
from flake8_qa_style.checkers.raw_checkers import FileStartChecker

from .checkers.raw_checkers._raw_checker import RawChecker
from .config import Config


def str_to_bool(string):
    return string.lower() in ('true', 'yes', 't', '1')


class QAStylePlugin(Plugin):
    name = 'flake8_qa_style'
    version = '1.1.0'

    visitors = [
        AnnotationVisitor,
        FunctionCallVisitor,
        AssertVisitor,
        WithVisitor,
    ]

    checkers = [
        FileStartChecker,
    ]

    def __init__(self, tree: ast.AST, filename: str, lines: list[str]):
        super().__init__(tree)
        self.filename = filename
        self.lines = lines

    def run(self):
        for checker_cls in self.checkers:
            checker = self._create_checker(checker_cls, filename=self.filename, lines=self.lines)
            checker.check()
            for error in checker.errors:
                yield self._error(error)

        for visitor_cls in self.visitors:
            visitor = self._create_visitor(visitor_cls, filename=self.filename)
            visitor.visit(self._tree)
            for error in visitor.errors:
                yield self._error(error)

    @classmethod
    def _create_checker(
        cls, checker_cls: Callable, filename: Optional[str] = None, lines: Optional[list[str]] = None
    ) -> RawChecker:
        kwargs = {}

        if filename is not None:
            kwargs['filename'] = filename
        if lines is not None:
            kwargs['lines'] = lines

        return checker_cls(**kwargs)

    @classmethod
    def _create_visitor(
        cls, visitor_cls: Callable, filename: Optional[str] = None
    ) -> Visitor:
        kwargs = {}

        if filename is not None:
            kwargs['filename'] = filename
        if cls.config is not None:
            kwargs['config'] = cls.config

        return visitor_cls(**kwargs)

    @classmethod
    def add_options(cls, option_manager: OptionManager):
        option_manager.add_option(
            '--skip-property-return-annotation',
            type=str,
            default='false',
            parse_from_config=True,
            help='Flag to skip return value annotation check. '
                 '(Default: False)',
        )

    @classmethod
    def parse_options_to_config(
        cls, option_manager: OptionManager, options: argparse.Namespace, args: List[str]
    ) -> Config:
        return Config(
            skip_property_return_annotation=str_to_bool(options.skip_property_return_annotation),
        )



import ast

from flake8_qa_style.checkers.node_visitors._node_visitor import NodeVisitor
from flake8_qa_style.errors import AssertSameObjectsForEquality, AssertWithConstant


class AssertVisitor(NodeVisitor):

    def visit_Assert(self, node: ast.Assert):
        if isinstance(node.test, ast.Compare):

            left = node.test.left
            right = node.test.comparators[0]

            # assert 1 == 2
            if isinstance(left, ast.Constant) and isinstance(right, ast.Constant):

                if left.value == right.value:
                    self.error_from_node(AssertSameObjectsForEquality, node)

            # assert var == var
            elif isinstance(left, ast.Name) and isinstance(right, ast.Name):
                if left.id == right.id:
                    self.error_from_node(AssertSameObjectsForEquality, node)

            # assert var == 'string' or assert var == 1
            elif isinstance(right, ast.Constant):
                self.error_from_node(AssertWithConstant, node)

            # assert self.foo == self.foo
            elif isinstance(left, ast.Attribute) and isinstance(right, ast.Attribute):
                if left.attr == right.attr:  # foo == foo
                    if (
                        isinstance(left.value, ast.Name) and isinstance(right.value, ast.Name)
                        and left.value.id == right.value.id
                    ):
                        self.error_from_node(AssertSameObjectsForEquality, node)



import ast
from typing import Union

from flake8_qa_style.checkers.node_visitors._node_visitor import NodeVisitor
from flake8_qa_style.errors import ArgAnnotationMissing, ReturnAnnotationMissing


class AnnotationVisitor(NodeVisitor):

    def check_annotation(self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):

        # exclude annotation check for dunder methods
        if node.name.startswith('__') and node.name.endswith('__'):
            return

        for arg in node.args.args:
            if arg.annotation is None and arg.arg != 'self' and arg.arg != 'cls':
                self.error_from_node(ArgAnnotationMissing, node,
                                     arg_name=arg.arg,
                                     func_name=node.name)

        if self.config.skip_property_return_annotation:
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Name) and decorator.id == 'property':
                    return

        for line in node.body:
            if isinstance(line, ast.Return):
                if node.returns is None:
                    self.error_from_node(ReturnAnnotationMissing, node,
                                         func_name=node.name)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.check_annotation(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self.check_annotation(node)



import ast
from typing import List

from flake8_plugin_utils import Error

from flake8_qa_style.checkers.node_visitors.function_call_visitor import (
    Checker,
    FunctionCallVisitor,
)
from flake8_qa_style.errors import Print


@FunctionCallVisitor.register_checker
class PrintChecker(Checker):

    def check(self, call_node: ast.Call, *args) -> List[Error]:
        if isinstance(call_node.func, ast.Name):
            if call_node.func.id == 'print' or call_node.func.id == 'pp':
                return [Print(call_node.lineno, call_node.col_offset)]
        return []



from .errors import (
    ArgAnnotationMissing,
    AssertSameObjectsForEquality,
    AssertWithConstant,
    NestedWith,
    NoBlankFileStart,
    Print,
    ReturnAnnotationMissing,
    SleepWithConstantArgument,
)



import ast
from abc import ABC, abstractmethod
from typing import List, Optional, Type

from flake8_qa_style.checkers.node_visitors._node_visitor import NodeVisitor


class Checker(ABC):
    @abstractmethod
    def check(self, call_node, context):
        pass


class Context:
    def __init__(self, sleep_function_name: Optional[str], is_imported_module_time: bool):
        self.sleep_function_name = sleep_function_name
        self.is_imported_module_time = is_imported_module_time


class FunctionCallVisitor(NodeVisitor):
    checkers: List[Checker] = []
    is_imported_module_time: bool = False
    sleep_function_name: Optional[str] = None

    @classmethod
    def register_checker(cls, checker: Type[Checker]):
        cls.checkers.append(checker())
        return checker

    @classmethod
    def deregister_all(cls):
        cls.checkers = []

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module == 'time':
            for name in node.names:
                if name.name == 'sleep':
                    self.sleep_function_name = name.name if not name.asname else name.asname

    def visit_Import(self, node: ast.ImportFrom):
        for name in node.names:
            if name.name == 'time':
                self.is_imported_module_time = True

    def visit_Call(self, node: ast.Call):
        context = Context(self.sleep_function_name, self.is_imported_module_time)
        try:
            for checker in self.checkers:
                self.errors.extend(checker.check(node, context))
        except Exception as e:
            print(f'Linter failed: checking {self.filename} with {checker.__class__}.\n'
                  f'Exception: {e}')



import ast

from flake8_qa_style.checkers.node_visitors._node_visitor import NodeVisitor
from flake8_qa_style.errors import NestedWith


class WithVisitor(NodeVisitor):

    def _check_nested_with(self, node: ast.With | ast.AsyncWith) -> None:
        is_async = isinstance(node, ast.AsyncWith)

        # Check for nested "with" of the same type (sync-in-sync or async-in-async)
        if len(node.body) == 1:
            inner = node.body[0]
            if is_async and isinstance(inner, ast.AsyncWith):
                self.error_from_node(NestedWith, node)
            elif not is_async and isinstance(inner, ast.With):
                self.error_from_node(NestedWith, node)

    def visit_With(self, node: ast.With) -> None:
        self._check_nested_with(node)
        self.generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        self._check_nested_with(node)
        self.generic_visit(node)



from .annotation_visitor import AnnotationVisitor
from .assert_visitor import AssertVisitor
from .call_checkers import PrintChecker, SleepChecker
from .function_call_visitor import FunctionCallVisitor
from .with_visitor import WithVisitor



from flake8_plugin_utils import assert_error, assert_not_error

from flake8_qa_style.checkers.node_visitors import WithVisitor
from flake8_qa_style.errors import NestedWith
from tests.test_helpers import assert_errors


def test_nested_with_sync_in_sync():
    code = """\
    with open('a'):
        with open('b'):
            pass
    """
    assert_error(WithVisitor, code, NestedWith)


def test_nested_with_async_in_async():
    code = """\
    async with open('a'):
        async with open('b'):
            pass
    """
    assert_error(WithVisitor, code, NestedWith)


def test_nested_with_sync_in_async():
    code = """\
    async with open('a'):
        with open('b'):
            pass
    """
    assert_not_error(WithVisitor, code)


def test_nested_with_async_in_sync():
    code = """\
    with open('a'):
        async with open('b'):
            pass
    """
    assert_not_error(WithVisitor, code)


def test_single_with_no_error():
    code = """\
    with open('a'):
        pass
    """
    assert_not_error(WithVisitor, code)


def test_single_async_with_no_error():
    code = """\
    async with open('a'):
        pass
    """
    assert_not_error(WithVisitor, code)


def test_combined_with_parenthesized():
    code = """\
    with (
        open('a'),
        open('b'),
    ):
        pass
    """
    assert_not_error(WithVisitor, code)


def test_combined_with_comma():
    code = """\
    with open('a'), open('b'):
        pass
    """
    assert_not_error(WithVisitor, code)

def test_nested_with_body_has_other_statements():
    code = """\
    with open('a'):
        x = 1
        with open('b'):
            pass
    """
    assert_not_error(WithVisitor, code)


def test_few_nested_with():
    code = """\
    with open('a'):
        with open('b'):
            with open('c'):
                pass
    """
    assert_errors(WithVisitor, code, [NestedWith, NestedWith])


def test_nested_with_single_item_each():
    code = """\
    with open('file') as f:
        with f.read() as data:
            pass
    """
    assert_error(WithVisitor, code, NestedWith)
